"""
Client for Commune modules.
"""

import asyncio
import json
from typing import Any

import aiohttp
import aiohttp.client_exceptions
import aiohttp.web_exceptions
from substrateinterface import Keypair  # type: ignore

from communex.errors import NetworkTimeoutError
from communex.key import check_ss58_address
from communex.types import Ss58Address

from ._protocol import create_method_endpoint, create_request_data
from ._signer import TESTING_MNEMONIC


class ModuleClient:
    host: str
    port: int
    key: Keypair

    def __init__(self, host: str, port: int, key: Keypair):
        self.host = host
        self.port = port
        self.key = key

    async def call(
            self,
            fn: str,
            target_key: Ss58Address,
            params: Any = {},
            timeout: int = 16,
    ) -> Any:
        serialized_data, headers = create_request_data(self.key, target_key, params)

        out = aiohttp.ClientTimeout(total=timeout)
        try:
            async with aiohttp.ClientSession(timeout=out) as session:
                async with session.post(
                    create_method_endpoint(self.host, self.port, fn),
                    json=json.loads(serialized_data),
                    headers=headers,
                ) as response:
                    match response.status:
                        case 200:
                            pass
                        case status_code:
                            response_j = await response.json()
                            raise Exception(
                                f"Unexpected status code: {status_code}, response: {response_j}")
                    match response.content_type:
                        case 'application/json':
                            result = await asyncio.wait_for(response.json(), timeout=timeout)
                            # TODO: deserialize result
                            return result
                        case _:
                            raise Exception(
                                f"Unknown content type: {response.content_type}")
        except asyncio.exceptions.TimeoutError as e:
            raise NetworkTimeoutError(
                f"The call took longer than the timeout of {timeout} second(s)").with_traceback(e.__traceback__)


if __name__ == "__main__":
    keypair = Keypair.create_from_mnemonic(
        TESTING_MNEMONIC
    )
    client = ModuleClient("localhost", 8000, keypair)
    ss58_address = check_ss58_address(keypair.ss58_address)
    result = asyncio.run(client.call("do_the_thing", ss58_address, {
                         "awesomness": 45, "extra": "hi"}))
    print(result)



from typing import Awaitable, Callable

from fastapi import Request, Response
from fastapi.responses import JSONResponse
from keylimiter import TokenBucketLimiter
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.types import ASGIApp
from pydantic_settings import BaseSettings


Callback = Callable[[Request], Awaitable[Response]]


class IpLimiterParams(BaseSettings):
    bucket_size: int = 15
    refill_rate: int = 1

    class config:
        env_prefix = "CONFIG_IP_LIMITER_"
        extra = "ignore"


class StakeLimiterParams(BaseSettings):
    epoch: int = 800
    cache_age: int = 600
    get_refill_per_epoch: Callable[[int], float] | None = None
    token_ratio: int = 1

    class config:
        env_prefix = "CONFIG_STAKE_LIMITER_"
        extra = "ignore"


class IpLimiterMiddleware(BaseHTTPMiddleware):
    def __init__(
            self,
            app: ASGIApp,
            params: IpLimiterParams | None,
    ):
        '''
        :param app: FastAPI instance
        :param limiter: KeyLimiter instance OR None

        If limiter is None, then a default TokenBucketLimiter is used with the following config:
        bucket_size=200, refill_rate=15
        '''
        super().__init__(app)

        # fallback to default limiter
        if not params:
            params = IpLimiterParams()
        self._limiter = TokenBucketLimiter(bucket_size=params.bucket_size, refill_rate=params.refill_rate)

    async def dispatch(self, request: Request, call_next: Callback) -> Response:
        assert request.client is not None, "request is invalid"
        assert request.client.host, "request is invalid."

        ip = request.client.host

        is_allowed = self._limiter.allow(ip)

        if not is_allowed:
            response = JSONResponse(
                status_code=429,
                headers={"X-RateLimit-Remaining": str(self._limiter.remaining(ip))},
                content={"error": "Rate limit exceeded"}
            )
            return response

        response = await call_next(request)

        return response



import time
from collections import OrderedDict
from collections.abc import MutableMapping
from threading import Lock
from typing import Callable, Generic, Iterator, TypeVar

K = TypeVar("K")
V = TypeVar("V")


class TTLDict(Generic[K, V], MutableMapping[K, V]):
    """
    A dictionary that expires its values after a given timeout.

    Time tracking is done with `time.monotonic_ns()`.

    Be careful, this doens't have and automatic cleanup of expired values yet.
    """

    ttl: int
    _values: OrderedDict[K, tuple[int, V]]
    _lock: Lock

    def __init__(self, ttl: int, _dict_type: type[OrderedDict[K, tuple[int, V]]] = OrderedDict):
        """
        Args:
            ttl: The timeout in seconds for the memoization.
        """
        self.ttl = ttl
        self._values = _dict_type()
        self._lock = Lock()

    @property
    def ttl_in_ns(self) -> int:
        return self.ttl * 1_000_000_000

    def __repr__(self) -> str:
        return f"<TTLDict@{id(self):#08x} ttl={self.ttl}>"

    def __is_expired(self, key: K) -> bool:
        expire_time, _ = self._values[key]
        return time.monotonic_ns() > expire_time

    def __remove_if_expired(self, key: K) -> bool:
        expired = self.__is_expired(key)
        if expired:
            del self._values[key]
        return expired

    def clean(self):
        with self._lock:
            for key in self._values.keys():
                removed = self.__remove_if_expired(key)
                if not removed:
                    # TODO: test cleanup optimization
                    break

    def __setitem__(self, key: K, value: V):
        with self._lock:
            expire_time = time.monotonic_ns() + self.ttl_in_ns
            self._values[key] = (expire_time, value)
            self._values.move_to_end(key)

    def __getitem__(self, key: K) -> V:
        with self._lock:
            self.__remove_if_expired(key)
            value = self._values[key][1]
            return value

    def __delitem__(self, key: K):
        with self._lock:
            del self._values[key]

    def __iter__(self) -> Iterator[K]:
        with self._lock:
            for key in self._values.keys():
                if not self.__remove_if_expired(key):
                    yield key

    def __len__(self) -> int:
        """
        Warning: this triggers a cleanup, and is O(n) in the number of items in
        the dict.
        """
        self.clean()

        # TODO: there is a race condition here.
        # Because we are not using RLock as we expect this crap to be used with async.
        # But I don't care. Be happy enough with with an "aproximate value".

        with self._lock:
            return len(self._values)

    def get_or_insert_lazy(self, key: K, fn: Callable[[], V]) -> V:
        """
        Gets the value for the given key, or inserts the value returned by the
        given function if the key is not present, returning it.
        """
        if key in self:
            return self[key]
        else:
            self[key] = fn()
            return self[key]


def __test():
    m: TTLDict[str, int] = TTLDict(1)

    m["a"] = 2

    print(m.get("a", default="missing"))
    print(m["a"])

    time.sleep(0.5)

    print(m.get("a", default="missing"))
    print(m["a"])

    time.sleep(1)

    print(m.get("a", default="missing"))
    try:
        print(m["a"])

    except KeyError:
        print("Key is not present :) yay")

    print(len(m))

    print()
    print()

    _counter = 0

    def generate():
        nonlocal _counter
        _counter += 1
        print(f"-> generated {_counter}")
        return _counter

    print("FIRST RUN")
    v = m.get_or_insert_lazy("a", generate)
    print(v)
    v = m.get_or_insert_lazy("a", generate)
    print(v)
    print()

    time.sleep(1.5)

    print("SECOND RUN")
    v = m.get_or_insert_lazy("a", generate)
    print(v)
    v = m.get_or_insert_lazy("a", generate)
    print(v)


if __name__ == "__main__":
    __test()



from communex.module import Module, endpoint

# prompt: str = 'sup?',
# model: str = 'gpt-3.5-turbo',
# presence_penalty: float = 0.0,
# frequency_penalty: float = 0.0,
# temperature: float = 0.9,
# max_tokens: int = 100,def endpoint(fn: Any):
# top_p: float = 1,
# choice_idx: int = 0,
# api_key: str = None,
# retry: bool = True,
# role: str = 'user',
# history: list = None,


class OpenAI(Module):
    @endpoint
    def generate(self, prompt: str, model: str = 'gpt-3.5-turbo'):
        print(f"Answering: `{prompt}` with model `{model}`")



"""
Common types for the communex module.
"""

from typing import NewType, TypedDict

Ss58Address = NewType("Ss58Address", str)
"""Substrate SS58 address.

The `SS58 encoded address format`_ is based on the Bitcoin Base58Check format,
but with a few modification specifically designed to suite Substrate-based
chains.

.. _SS58 encoded address format:
    https://docs.substrate.io/reference/address-formats/
"""


# TODO: replace with dataclasses

# == Burn related
MinBurn = NewType("MinBurn", int)
MaxBurn = NewType("MaxBurn", int)
BurnConfig = NewType("BurnConfig", dict[MinBurn, MaxBurn])


class NetworkParams(TypedDict):
    max_allowed_modules: int
    max_registrations_per_block: int
    max_name_length: int
    min_name_length: int
    min_burn: int
    max_burn: int
    min_weight_stake: int
    max_allowed_subnets: int
    curator: Ss58Address
    proposal_cost: int
    proposal_expiration: int
    subnet_stake_threshold: int
    general_subnet_application_cost: int
    floor_founder_share: int
    floor_delegation_fee: int
    max_allowed_weights: int


class SubnetParamsMaps(TypedDict):
    netuid_to_founder: dict[int, Ss58Address]
    netuid_to_founder_share: dict[int, int]
    netuid_to_immunity_period: dict[int, int]
    netuid_to_incentive_ratio: dict[int, int]
    netuid_to_max_allowed_uids: dict[int, int]
    netuid_to_max_allowed_weights: dict[int, int]
    netuid_to_min_allowed_weights: dict[int, int]
    netuid_to_max_weight_age: dict[int, int]
    netuid_to_min_stake: dict[int, int]
    netuid_to_name: dict[int, str]
    netuid_to_tempo: dict[int, int]
    netuid_to_trust_ratio: dict[int, int]
    netuid_to_bonds_ma: dict[int, int]
    netuid_to_maximum_set_weight_calls_per_epoch: dict[int, int]
    netuid_to_target_registrations_per_interval: dict[int, int]
    netuid_to_target_registrations_interval: dict[int, int]
    netuid_to_emission: dict[int, int]
    netuid_to_max_registrations_per_interval: dict[int, int]
    netuid_to_vote_mode: dict[int, dict[str, str]]
    netuid_to_adjustment_alpha: dict[int, int]


class SubnetParams(TypedDict):
    founder: Ss58Address
    founder_share: int
    immunity_period: int
    incentive_ratio: int
    max_allowed_uids: int
    max_allowed_weights: int
    min_allowed_weights: int
    max_weight_age: int
    min_stake: int
    name: str
    tempo: int
    trust_ratio: int
    bonds_ma: int | None
    maximum_set_weight_calls_per_epoch: int | None
    target_registrations_per_interval: int
    target_registrations_interval: int
    max_registrations_per_interval: int
    vote_mode: str
    adjustment_alpha: int


# redundant "TypedDict" inheritance because of pdoc warns.
# see https://github.com/mitmproxy/pdoc/blob/26d40827ddbe1658e8ac46cd092f17a44cf0287b/pdoc/doc.py#L691-L692
class SubnetParamsWithEmission(SubnetParams, TypedDict):
    """SubnetParams with emission field."""

    emission: int
    """Subnet emission percentage (0-100).
    """


class ModuleInfo(TypedDict):
    uid: int
    key: Ss58Address
    name: str
    address: str  # "<ip>:<port>"
    emission: int
    incentive: int
    dividends: int
    stake_from: list[tuple[str, int]]  # TODO: type key with Ss58Address
    regblock: int  # block number
    last_update: int  # block number
    stake: int
    delegation_fee: int
    metadata: str


class ModuleInfoWithBalance(ModuleInfo):
    balance: int


class ModuleInfoWithOptionalBalance(ModuleInfo):
    balance: int | None



from asyncio import Lock
from math import ceil, floor
from time import monotonic
from typing import Callable

from substrateinterface.utils.ss58 import ss58_decode  # type: ignore

from communex._common import get_node_url
from communex.balance import to_nano
from communex.client import CommuneClient


def local_keys_to_stakedbalance(netuid: list[int]) -> dict[str, int]:
    url = get_node_url()
    client = CommuneClient(url)
    total_stake: dict[str, int] = {}
    for uid in netuid:

        query_all = client.query_batch_map(
            {
                "SubspaceModule": [("StakeFrom", [uid])],
            })

        staketo_map = query_all["StakeFrom"]

        for key, value in staketo_map.items():
            key_stake = sum(stake for _, stake in value)
            total_stake.setdefault(key, 0)
            total_stake[key] += key_stake

    return total_stake


def calls_per_epoch(stake: int, multiplier: int = 1) -> float:
    """Gives how many requests per epoch a stake can make"""
    max_ratio = 4
    base_ratio = 89
    if multiplier <= 1 / max_ratio:
        raise ValueError(
            f"Given multiplier {multiplier} would set 0 tokens for all stakes"
        )

    def mult_2(x: int) -> int:
        return x * 2

    # 10x engineer switch case
    match stake:
        case _ if stake < to_nano(10_000):
            return 0
        case _ if stake < to_nano(500_000):
            return base_ratio * multiplier
        case _:
            return mult_2(base_ratio) * multiplier


def build_keys_refill_rate(
    netuid: list[int] | None,
    get_refill_rate: Callable[[int], float] = calls_per_epoch
):
    if netuid is None:
        empty_dict: dict[str, float] = {}
        return empty_dict
    key_to_stake = local_keys_to_stakedbalance(netuid)
    key_to_ratio = {ss58_decode(key): get_refill_rate(stake) for key, stake in key_to_stake.items()}
    return key_to_ratio


class StakeLimiter():
    def __init__(
        self,
        subnets_whitelist: list[int] | None,
        time_func: Callable[[], float] = monotonic,
        epoch: int = 800,
        get_refill_rate: Callable[[int], float] | None = None,
        max_cache_age: int = 600,
    ):
        self._time = time_func
        if get_refill_rate is None:
            get_refill_rate = calls_per_epoch

        self.refiller_function = get_refill_rate
        self._lock = Lock()

        self.buckets: dict[str, tuple[float, float]] = {}

        self.whitelist = subnets_whitelist
        self.key_ratio = build_keys_refill_rate(
            netuid=subnets_whitelist, get_refill_rate=self.refiller_function
        )
        self.key_ratio_age = monotonic()
        self.max_cache_age = max_cache_age

        self.epoch = epoch

    async def _get_key_refresh_ratio(self, key: str) -> float:
        # Every access to key_ratio should pass through here so we
        # can update the cache when its too old.

        if not self.whitelist:
            return 1000
        if monotonic() - self.key_ratio_age > self.max_cache_age:
            self.key_ratio = build_keys_refill_rate(
                netuid=self.whitelist,
                get_refill_rate=self.refiller_function,
            )
            self.key_ratio_age = monotonic()
        ratio = self.key_ratio.get(key, 0)
        if ratio == 0:
            return 0
        return ratio / self.epoch

    async def _get_key_ratio_per_epoch(self, key: str) -> float:
        return await self._get_key_refresh_ratio(key) * self.epoch

    async def allow(self, key: str) -> bool:
        if not self.whitelist:
            # basically test mode that disables validation
            return True
        async with self._lock:
            return await self._allow(key)

    async def _allow(self, key: str) -> bool:
        tokens = await self._remaining(key)
        if tokens >= 1:
            self._set_tokens(key, tokens - 1)
            return True
        return False

    def limit(self, key: str) -> int:
        key_rate = self.key_ratio.get(key, 0)
        tokens = max(1, key_rate)
        return int(tokens)

    async def remaining(self, key: str) -> int:
        async with self._lock:
            remaining = await self._remaining(key)
        return floor(remaining)

    async def _remaining(self, key: str) -> float:
        await self._refill(key)
        tokens, _ = self.buckets.get(key, (0, 0))
        return tokens

    async def retry_after(self, key: str) -> int:
        async with self._lock:
            return await self._retry_after(key)

    async def _retry_after(self, key: str) -> int:
        tokens = await self._remaining(key)
        if tokens >= 1:
            return 0
        key_rate = await self._get_key_refresh_ratio(key)
        if key_rate > 0:
            return ceil(1 / key_rate)
        else:
            return self.max_cache_age
        
    async def _refill(self, key: str) -> None:
        bucket = self.buckets.get(key)
        if bucket is None:  # type: ignore
            await self._fill(key)
            return

        filled_bucket = self.buckets.get(key)
        assert filled_bucket  # type: ignore
        tokens, last_seen = filled_bucket

        key_rate = await self._get_key_refresh_ratio(key)
        new_tokens = floor((monotonic() - last_seen) * key_rate)

        if new_tokens <= 0:
            return

        tokens = min(tokens + new_tokens, max(self.key_ratio.values(), default=0))  # sink overflow

        self._set_tokens(key, tokens)  # has race conditions in multi-threaded environments

    async def _fill(self, key: str) -> None:
        # starts with at least 1 token
        ratio = await self._get_key_ratio_per_epoch(key)
        tokens = max(1, ratio)
        self._set_tokens(key, int(tokens))

    def _set_tokens(self, key: str, tokens: float) -> None:
        self.buckets[key] = (tokens, monotonic())



from typing import Protocol, Any, Sequence
from abc import abstractmethod
import json
import re
from datetime import datetime, timezone

from fastapi import Request, Response
from fastapi.routing import APIRoute
from fastapi.responses import JSONResponse
from keylimiter import TokenBucketLimiter
from substrateinterface import Keypair  # type: ignore
import starlette.datastructures
from functools import partial

from communex.module import _signer as signer
from communex.types import Ss58Address
from communex.module._util import log, log_reffusal, json_error, try_ss58_decode
from communex.module._rate_limiters._stake_limiter import StakeLimiter
from communex.module._rate_limiters.limiters import StakeLimiterParams, IpLimiterParams
from communex.util.memo import TTLDict
from communex.module._util import make_client
from communex._common import get_node_url


HEX_PATTERN = re.compile(r"^[0-9a-fA-F]+$")
def is_hex_string(string: str):
    return bool(HEX_PATTERN.match(string))


def parse_hex(hex_str: str) -> bytes:
    if hex_str[0:2] == "0x":
        return bytes.fromhex(hex_str[2:])
    else:
        return bytes.fromhex(hex_str)
    

class AbstractVerifier(Protocol):
    @abstractmethod
    async def verify(self, request: Request) -> JSONResponse | None:
        """Please dont mutate the request D:"""
        ...

class StakeLimiterVerifier(AbstractVerifier):
    def __init__(
        self,
        subnets_whitelist: list[int] | None,
        params_: StakeLimiterParams | None,
    ):
        self.subnets_whitelist = subnets_whitelist
        self.params_ = params_

    async def verify(self, request: Request):
        if not self.params_:
            params = StakeLimiterParams()
        else:
            params = self.params_
        limiter = StakeLimiter(
            self.subnets_whitelist, 
            epoch=params.epoch, 
            max_cache_age=params.cache_age,
            get_refill_rate=params.get_refill_per_epoch,
            )

        if request.client is None:
            response = JSONResponse(
                status_code=401,
                content={
                    "error": "Address should be present in request"
                }
            )
            return response
        
        key = request.headers.get('x-key')
        if not key:
            response = JSONResponse(
                status_code=401,
                content={"error": "Valid X-Key not provided on headers"}
                )
            return response

        is_allowed = await limiter.allow(key)

        if not is_allowed:
            response = JSONResponse(
                status_code=429, 
                headers={"X-RateLimit-TryAfter": f"{str(await limiter.retry_after(key))} seconds"},
                content={"error": "Rate limit exceeded"}
                )
            return response
        return None


class ListVerifier(AbstractVerifier):
    def __init__(
            self,
            blacklist: list[Ss58Address] | None,
            whitelist: list[Ss58Address] | None,
            ip_blacklist: list[str] | None,
    ):
        self.blacklist = blacklist
        self.whitelist = whitelist
        self.ip_blacklist = ip_blacklist
    
    async def verify(self, request: Request) -> JSONResponse | None:
        key = request.headers.get("x-key")
        if not key:
            reason = "Missing header: X-Key"
            log(f"INFO: refusing module request because: {reason}")
            return json_error(400, "Missing header: X-Key")

        ss58 = try_ss58_decode(key)
        if ss58 is None:
            reason = "Caller key could not be decoded into a ss58address"
            log_reffusal(key, reason)
            return json_error(400, reason)
        if request.client is None:
            return json_error(400, "Address should be present in request")
        if self.blacklist and ss58 in self.blacklist:
            return json_error(403, "You are blacklisted")
        if self.ip_blacklist and request.client.host in self.ip_blacklist:
            return json_error(403, "Your IP is blacklisted")
        if self.whitelist and ss58 not in self.whitelist:
            return json_error(403, "You are not whitelisted")
        return None
        

class IpLimiterVerifier(AbstractVerifier):
    def __init__(
            self,
            params: IpLimiterParams | None,
    ):
        '''
        :param limiter: KeyLimiter instance OR None

        If limiter is None, then a default TokenBucketLimiter is used with the following config:
        bucket_size=200, refill_rate=15
        '''

        # fallback to default limiter
        if not params:
            params = IpLimiterParams()
        self._limiter = TokenBucketLimiter(bucket_size=params.bucket_size, refill_rate=params.refill_rate)

    async def verify(self, request: Request):
        assert request.client is not None, "request is invalid"
        assert request.client.host, "request is invalid."

        ip = request.client.host

        is_allowed = self._limiter.allow(ip)

        if not is_allowed:
            response = JSONResponse(
                status_code=429, 
                headers={"X-RateLimit-Remaining": str(self._limiter.remaining(ip))},
                content={"error": "Rate limit exceeded"}
                )
            return response
        return None
    

class InputHandlerVerifier(AbstractVerifier):
        def __init__(
            self,
            subnets_whitelist: list[int] | None,
            module_key: Ss58Address,
            request_staleness: int,
            blockchain_cache: TTLDict[str, list[Ss58Address]],
            host_key: Keypair,
            use_testnet: bool,
        ):
            self.subnets_whitelist = subnets_whitelist
            self.module_key = module_key
            self.request_staleness = request_staleness
            self.blockchain_cache = blockchain_cache
            self.host_key = host_key
            self.use_testnet = use_testnet


        async def verify(self, request: Request):
            body = await request.body()

            # TODO: we'll replace this by a Result ADT :)
            match self._check_inputs(request, body, self.module_key):
                case (False, error):
                    return error
                case (True, _):
                    pass
            
            body_dict: dict[str, dict[str, Any]] = json.loads(body)
            timestamp = body_dict['params'].get("timestamp", None)
            legacy_timestamp = request.headers.get("X-Timestamp", None)
            try:
                timestamp_to_use = timestamp if not legacy_timestamp else legacy_timestamp
                request_time = datetime.fromisoformat(timestamp_to_use)
            except Exception:
                return JSONResponse(status_code=400, content={"error": "Invalid ISO timestamp given"})
            if (datetime.now(timezone.utc) - request_time).total_seconds() > self.request_staleness:
                return JSONResponse(status_code=400, content={"error": "Request is too stale"})
            return None


        def _check_inputs(
                self, request: Request, 
                body: bytes, 
                module_key: Ss58Address
            ):
            required_headers = ["x-signature", "x-key", "x-crypto"]
            optional_headers = ["x-timestamp"]

            # TODO: we'll replace this by a Result ADT :)
            match self._get_headers_dict(request.headers, required_headers, optional_headers):
                case (False, error):
                    return (False, error)
                case (True, headers_dict):
                    pass

            # TODO: we'll replace this by a Result ADT :)
            match self._check_signature(headers_dict, body, module_key):
                case (False, error):
                    return (False, error)
                case (True, _):
                    pass

            # TODO: we'll replace this by a Result ADT :)
            match self._check_key_registered(
                self.subnets_whitelist,
                headers_dict,
                self.blockchain_cache,
                self.host_key,
                self.use_testnet,
            ):
                case (False, error):
                    return (False, error)
                case (True, _):
                    pass

            return (True, None)
        
        def _get_headers_dict(
                self,
                headers: starlette.datastructures.Headers,
                required: list[str],
                optional: list[str],
            ):
            headers_dict: dict[str, str] = {}
            for required_header in required:
                value = headers.get(required_header)
                if not value:
                    code = 400
                    return False, json_error(code, f"Missing header: {required_header}")
                headers_dict[required_header] = value
            for optional_header in optional:
                value = headers.get(optional_header)
                if value:
                    headers_dict[optional_header] = value

            return True, headers_dict


        def _check_signature(
                self,
                headers_dict: dict[str, str],
                body: bytes,
                module_key: Ss58Address
            ):
            key = headers_dict["x-key"]
            signature = headers_dict["x-signature"]
            crypto = int(headers_dict["x-crypto"])  # TODO: better handling of this

            if not is_hex_string(key):
                reason = "X-Key should be a hex value"
                log_reffusal(key, reason)
                return (False, json_error(400, reason))
            try:
                signature = parse_hex(signature)
            except Exception:
                reason = "Signature sent is not a valid hex value"
                log_reffusal(key, reason)
                return False, json_error(400, reason)
            try:
                key = parse_hex(key)
            except Exception:
                reason = "Key sent is not a valid hex value"
                log_reffusal(key, reason)
                return False, json_error(400, reason)
            # decodes the key for better logging
            key_ss58 = try_ss58_decode(key)
            if key_ss58 is None:
                reason = "Caller key could not be decoded into a ss58address"
                log_reffusal(key.decode(), reason)
                return (False, json_error(400, reason))

            timestamp = headers_dict.get("x-timestamp")
            legacy_verified = False
            if timestamp:
                # tries to do a legacy verification
                json_body = json.loads(body)
                json_body["timestamp"] = timestamp
                stamped_body = json.dumps(json_body).encode()
                legacy_verified = signer.verify(key, crypto, stamped_body, signature)

            verified = signer.verify(key, crypto, body, signature)
            if not verified and not legacy_verified:
                reason = "Signature doesn't match"
                log_reffusal(key_ss58, reason)
                return (False, json_error(401, "Signatures doesn't match"))

            body_dict: dict[str, dict[str, Any]] = json.loads(body)
            target_key = body_dict['params'].get("target_key", None)
            if not target_key or target_key != module_key:
                reason = "Wrong target_key in body"
                log_reffusal(key_ss58, reason)
                return (False, json_error(401, "Wrong target_key in body"))

            return (True, None)

        def _check_key_registered(
            self,
            subnets_whitelist: list[int] | None,
            headers_dict: dict[str, str],
            blockchain_cache: TTLDict[str, list[Ss58Address]],
            host_key: Keypair,
            use_testnet: bool,
        ):
            key = headers_dict["x-key"]
            if not is_hex_string(key):
                return (False, json_error(400, "X-Key should be a hex value"))
            key = parse_hex(key)

            # TODO: checking for key being registered should be smarter
            # e.g. query and store all registered modules periodically.

            ss58 = try_ss58_decode(key)
            if ss58 is None:
                reason = "Caller key could not be decoded into a ss58address"
                log_reffusal(key.decode(), reason)
                return (False, json_error(400, reason))

            # If subnets whitelist is specified, checks if key is registered in one
            # of the given subnets
            
            allowed_subnets: dict[int, bool] = {}
            caller_subnets: list[int] = []
            if subnets_whitelist is not None:
                def query_keys(subnet: int):
                    try:
                        node_url = get_node_url(None, use_testnet=use_testnet)
                        client = make_client(node_url)  # TODO: get client from outer context
                        return [*client.query_map_key(subnet).values()]
                    except Exception:
                        log(
                            "WARNING: Could not connect to a blockchain node"
                        )
                        return_list: list[Ss58Address] = []
                        return return_list
                
                # TODO: client pool for entire module server
                
                got_keys = False
                no_keys_reason = (
                    "Miner could not connect to a blockchain node "
                    "or there is no key registered on the subnet(s) {} "
                )
                for subnet in subnets_whitelist:
                    get_keys_on_subnet = partial(query_keys, subnet)
                    cache_key = f"keys_on_subnet_{subnet}"
                    keys_on_subnet = blockchain_cache.get_or_insert_lazy(
                        cache_key, get_keys_on_subnet
                    )
                    if len(keys_on_subnet) == 0:
                        reason = no_keys_reason.format(subnet)
                        log(f"WARNING: {reason}")
                    else:
                        got_keys = True

                    if host_key.ss58_address not in keys_on_subnet:
                        log(
                            f"WARNING: This miner is deregistered on subnet {subnet}"
                        )
                    else:
                        allowed_subnets[subnet] = True
                    if ss58 in keys_on_subnet:
                        caller_subnets.append(subnet)
                if not got_keys:
                    return False, json_error(503, no_keys_reason.format(subnets_whitelist))
                if not allowed_subnets:
                    log("WARNING: Miner is not registered on any subnet")
                    return False, json_error(403, "Miner is not registered on any subnet")
                
                # searches for a common subnet between caller and miner
                # TODO: use sets
                allowed_subnets = {
                    subnet: allowed for subnet, allowed in allowed_subnets.items() if (
                        subnet in caller_subnets
                        )
                    }
                if not allowed_subnets:
                    reason = "Caller key is not registered in any subnet that the miner is"
                    log_reffusal(ss58, reason)
                    return False, json_error(
                        403, reason
                    )
            else:
                # accepts everything
                pass
            
            return (True, None)


def build_route_class(
    verifiers: Sequence[AbstractVerifier]
) -> type[APIRoute]:

    class CheckListsRoute(APIRoute):
        def get_route_handler(self):
            original_route_handler = super().get_route_handler()

            async def custom_route_handler(request: Request) -> Response | JSONResponse:
                if not request.url.path.startswith('/method'):
                    unhandled_response: Response = await original_route_handler(request)
                    return unhandled_response
                for verifier in verifiers:
                    response = await verifier.verify(request)
                    if response is not None:
                        return response
                    
                original_response: Response = await original_route_handler(request)
                return original_response
            
            return custom_route_handler

    return CheckListsRoute




from threading import Lock
from types import TracebackType
from typing import ContextManager, Generic, TypeVar

T = TypeVar("T")


class MutexBox(Generic[T], ContextManager[T]):
    _mutex: Lock
    _value: T

    def __init__(self, value: T):
        self._mutex = Lock()
        self._value = value

    def __enter__(self) -> T:
        self._mutex.acquire()
        return self._value

    def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None):
        self._mutex.release()



import time

from substrateinterface import Keypair  # type: ignore

from communex._common import get_node_url
from communex.client import CommuneClient
from communex.compat.key import classic_load_key
from communex.types import Ss58Address

# The netuid of the general subnet.
GENERAL_NETUID = 0


def keys_to_uids(keys: dict[int, Ss58Address], target_keys: list[Ss58Address]) -> list[int]:
    return [k for k, v in keys.items() if v in target_keys]


def validaiton(client: CommuneClient, key: Keypair):
    # Query the modules, nominated by DAO.
    keys = client.query_map_key(netuid=GENERAL_NETUID)
    legit_whitelist = client.query_map(
        "LegitWhitelist", params=[], extract_value=False)["LegitWhitelist"]
    # TODO: test on production query, to verify the format, now I am going to assume it is a dict.
    target_keys = list(legit_whitelist.keys())
    uids_map = keys_to_uids(keys, target_keys)
    # WIP, rigt now assuming same weights
    weights = [1 for _ in uids_map]

    client.vote(key, uids=uids_map, weights=weights, netuid=GENERAL_NETUID)


def main(client: CommuneClient, key: Keypair,):
    while True:
        validaiton(client, key)
        time.sleep(1200)


if __name__ == "__main__":
    client = CommuneClient(get_node_url())
    key = classic_load_key("foo")
    main(client, key)



"""
Compatibility layer for the *classic* `commune` library.

Note:
    Anywhere you see **classic** in this docs, it means that the class/method/function/thing refers to or interacts with the `classic commune library`_.

.. _classic commune library:
    https://github.com/commune-ai/commune
"""



import random
from enum import Enum
import re
from typing import Mapping, TypeVar

from pydantic_settings import BaseSettings, SettingsConfigDict

from communex.balance import from_nano

IPFS_REGEX = re.compile(r"^Qm[1-9A-HJ-NP-Za-km-z]{44}$")

class ComxSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="COMX_")
    # TODO: improve node lists
    NODE_URLS: list[str] = [
        "wss://commune-api-node-0.communeai.net",
        "wss://commune-api-node-1.communeai.net",
        "wss://commune-api-node-2.communeai.net",
        "wss://commune-api-node-3.communeai.net",
        "wss://commune-api-node-4.communeai.net",
        "wss://commune-api-node-5.communeai.net",
        "wss://commune-api-node-6.communeai.net",
        "wss://commune-api-node-7.communeai.net",
        "wss://commune-api-node-8.communeai.net",
        "wss://commune-api-node-9.communeai.net",
        "wss://commune-api-node-10.communeai.net",
        "wss://commune-api-node-11.communeai.net",
        "wss://commune-api-node-12.communeai.net",
        "wss://commune-api-node-13.communeai.net",
        "wss://commune-api-node-14.communeai.net",
        "wss://commune-api-node-15.communeai.net",
        "wss://commune-api-node-16.communeai.net",
        "wss://commune-api-node-17.communeai.net",
        "wss://commune-api-node-18.communeai.net",
        "wss://commune-api-node-19.communeai.net",
        "wss://commune-api-node-20.communeai.net",
        "wss://commune-api-node-21.communeai.net",
        "wss://commune-api-node-22.communeai.net",
        "wss://commune-api-node-23.communeai.net",
        "wss://commune-api-node-24.communeai.net",
        "wss://commune-api-node-25.communeai.net",
        "wss://commune-api-node-26.communeai.net",
        "wss://commune-api-node-27.communeai.net",
        "wss://commune-api-node-28.communeai.net",
        "wss://commune-api-node-29.communeai.net",
        "wss://commune-api-node-30.communeai.net",
        "wss://commune-api-node-31.communeai.net",
    ]
    TESTNET_NODE_URLS: list[str] = ["wss://testnet-commune-api-node-0.communeai.net"]


def get_node_url(
    comx_settings: ComxSettings | None = None, *, use_testnet: bool = False
) -> str:
    comx_settings = comx_settings or ComxSettings()
    match use_testnet:
        case True:
            node_url = random.choice(comx_settings.TESTNET_NODE_URLS)
        case False:
            node_url = random.choice(comx_settings.NODE_URLS)
    return node_url


def get_available_nodes(
    comx_settings: ComxSettings | None = None, *, use_testnet: bool = False
) -> list[str]:
    comx_settings = comx_settings or ComxSettings()

    match use_testnet:
        case True:
            node_urls = comx_settings.TESTNET_NODE_URLS
        case False:
            node_urls = comx_settings.NODE_URLS
    return node_urls


class BalanceUnit(str, Enum):
    joule = "joule"
    j = "j"
    nano = "nano"
    n = "n"


def format_balance(balance: int, unit: BalanceUnit = BalanceUnit.nano) -> str:
    """
    Formats a balance.
    """

    match unit:
        case BalanceUnit.nano | BalanceUnit.n:
            return f"{balance}"
        case BalanceUnit.joule | BalanceUnit.j:
            in_joules = from_nano(balance)
            round_joules = round(in_joules, 4)
            return f"{round_joules:,} J"


K = TypeVar("K")
V = TypeVar("V")
Z = TypeVar("Z")


def intersection_update(base: dict[K, V], update: dict[K, Z]) -> Mapping[K, V | Z]:
    """
    Update a dictionary with another dictionary, but only with keys that are already present.
    """
    updated = {k: update[k] for k in base if k in update}
    return updated




from typing import TypedDict

from communex.types import Ss58Address


class CommuneStorageDict(TypedDict):
    data: str
    encrypted: bool
    timestamp: int


class CommuneKeyDict(TypedDict):
    crypto_type: int
    seed_hex: str
    derive_path: str | None
    path: str
    public_key: str
    ss58_format: int
    ss58_address: Ss58Address
    private_key: str
    mnemonic: str



import re
from typing import Any, TypeVar

from communex.client import CommuneClient
from communex.key import check_ss58_address
from communex.types import (
    ModuleInfoWithOptionalBalance,
    NetworkParams,
    Ss58Address,
    SubnetParamsWithEmission,
    SubnetParamsMaps,
)

IPFS_REGEX = re.compile(r"^Qm[1-9A-HJ-NP-Za-km-z]{44}$")

T = TypeVar("T")

def get_map_modules(
    client: CommuneClient,
    netuid: int = 0,
    include_balances: bool = False,
) -> dict[str, ModuleInfoWithOptionalBalance]:
    """
    Gets all modules info on the network
    """

    request_dict: dict[Any, Any] = {
        "SubspaceModule": [
            ("StakeFrom", [netuid]),
            ("Keys", [netuid]),
            ("Name", [netuid]),
            ("Address", [netuid]),
            ("RegistrationBlock", [netuid]),
            ("DelegationFee", [netuid]),
            ("Emission", []),
            
            ("Incentive", []),
            ("Dividends", []),
            ("LastUpdate", []),
            ("Metadata", [netuid]),
        ],
    }
    if include_balances:
        request_dict["System"] = [("Account", [])]

    bulk_query = client.query_batch_map(request_dict)
   
    (
        ss58_to_stakefrom,
        uid_to_key,
        uid_to_name,
        uid_to_address,
        uid_to_regblock,
        ss58_to_delegationfee,
        uid_to_emission,
        uid_to_incentive,
        uid_to_dividend,
        uid_to_lastupdate,
        ss58_to_balances,
        uid_to_metadata,
    ) = (
        bulk_query.get("StakeFrom", {}),
       bulk_query.get("Keys", {}),
        bulk_query["Name"],
        bulk_query["Address"],
        bulk_query["RegistrationBlock"],
        bulk_query["DelegationFee"],
        bulk_query["Emission"],
        bulk_query["Incentive"],
        bulk_query["Dividends"],
        bulk_query["LastUpdate"],
        bulk_query.get("Account", {}),
        bulk_query.get("Metadata", {}),
    )

    result_modules: dict[str, ModuleInfoWithOptionalBalance] = {}

    for uid, key in uid_to_key.items():
        key = check_ss58_address(key)

        name = uid_to_name[uid]
        address = uid_to_address[uid]
        emission = uid_to_emission[netuid][uid]
        incentive = uid_to_incentive[netuid][uid]
        dividend = uid_to_dividend[netuid][uid]
        regblock = uid_to_regblock[uid]
        stake_from = ss58_to_stakefrom.get(key, [])
        last_update = uid_to_lastupdate[netuid][uid]
        delegation_fee = ss58_to_delegationfee.get(
            key, 20
        )  # 20% default delegation fee
        metadata = uid_to_metadata.get(uid, None)

        balance = None
        if include_balances and ss58_to_balances is not None:  # type: ignore
            balance_dict = ss58_to_balances.get(key, None)
            if balance_dict is not None:
                assert isinstance(balance_dict["data"], dict)
                balance = balance_dict["data"]["free"]
            else:
                balance = 0
        stake = sum(stake for _, stake in stake_from)

        module: ModuleInfoWithOptionalBalance = {
            "uid": uid,
            "key": key,
            "name": name,
            "address": address,
            "emission": emission,
            "incentive": incentive,
            "dividends": dividend,
            "stake_from": stake_from,
            "regblock": regblock,
            "last_update": last_update,
            "balance": balance,
            "stake": stake,
            "delegation_fee": delegation_fee,
            "metadata": metadata,
        }

        result_modules[key] = module
    return result_modules

def to_snake_case(d: dict[str, T]) -> dict[str, T]:
    """
    Converts a dictionary with camelCase keys to snake_case keys
    """
    def snakerize(camel: str) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', camel).lower()
    snaked: dict[str, T] = {snakerize(k): v for k, v in d.items()}
    return snaked


def get_map_subnets_params(
    client: CommuneClient, block_hash: str | None = None
) -> dict[int, SubnetParamsWithEmission]:
    """
    Gets all subnets info on the network
    """

    bulk_query = client.query_batch_map(
        {
            "SubspaceModule": [
                ("ImmunityPeriod", []),
                ("MinAllowedWeights", []),
                ("MaxAllowedWeights", []),
                ("MinStake", []),
                ("SubnetEmission", []),
                ("Tempo", []),
                ("MaxAllowedUids", []),
                ("TargetRegistrationsInterval", []),
                ("TargetRegistrationsPerInterval", []),
                ("MaxRegistrationsPerInterval", []),
                ("Founder", []),
                ("FounderShare", []),
                ("IncentiveRatio", []),
                ("TrustRatio", []),
                ("SubnetNames", []),
                ("MaxWeightAge", []),
                ("BondsMovingAverage", []),
                ("MaximumSetWeightCallsPerEpoch", []),
                ("AdjustmentAlpha", []),
            ],
                "GovernanceModule": [
                    ("SubnetGovernanceConfig", []),
                ]
        },
        block_hash,
    )
    subnet_maps: SubnetParamsMaps = {
        "netuid_to_emission": bulk_query["SubnetEmission"],
        "netuid_to_tempo": bulk_query["Tempo"],
        "netuid_to_immunity_period": bulk_query["ImmunityPeriod"],
        "netuid_to_min_allowed_weights": bulk_query["MinAllowedWeights"],
        "netuid_to_max_allowed_weights": bulk_query["MaxAllowedWeights"],
        "netuid_to_max_allowed_uids": bulk_query["MaxAllowedUids"],
        "netuid_to_min_stake": bulk_query["MinStake"],
        "netuid_to_founder": bulk_query["Founder"],
        "netuid_to_founder_share": bulk_query["FounderShare"],
        "netuid_to_incentive_ratio": bulk_query["IncentiveRatio"],
        "netuid_to_trust_ratio": bulk_query["TrustRatio"],
        "netuid_to_name": bulk_query["SubnetNames"],
        "netuid_to_max_weight_age": bulk_query["MaxWeightAge"],
        "netuid_to_vote_mode": bulk_query["SubnetGovernanceConfig"],
        "netuid_to_adjustment_alpha": bulk_query["AdjustmentAlpha"],
        "netuid_to_bonds_ma": bulk_query.get("BondsMovingAverage", {}),
        "netuid_to_maximum_set_weight_calls_per_epoch": bulk_query.get("MaximumSetWeightCallsPerEpoch", {}),
        "netuid_to_target_registrations_per_interval": bulk_query.get("TargetRegistrationsPerInterval", {}),
        "netuid_to_target_registrations_interval": bulk_query.get("TargetRegistrationsInterval", {}),
        "netuid_to_max_registrations_per_interval": bulk_query.get("MaxRegistrationsPerInterval", {}),
    }
    result_subnets: dict[int, SubnetParamsWithEmission] = {}

    default_target_registrations_interval = 200
    default_target_registrations_per_interval = int(default_target_registrations_interval / 2)
    default_max_registrations_per_interval = 42
    for netuid, name in subnet_maps["netuid_to_name"].items():

        subnet: SubnetParamsWithEmission = {
            "name": name,
            "founder": subnet_maps["netuid_to_founder"][netuid],
            "founder_share": subnet_maps["netuid_to_founder_share"][netuid],
            "immunity_period": subnet_maps["netuid_to_immunity_period"][netuid],
            "incentive_ratio": subnet_maps["netuid_to_incentive_ratio"][netuid],
            "max_allowed_uids": subnet_maps["netuid_to_max_allowed_uids"][netuid],
            "max_allowed_weights": subnet_maps["netuid_to_max_allowed_weights"][netuid],
            "min_allowed_weights": subnet_maps["netuid_to_min_allowed_weights"][netuid],
            "min_stake": subnet_maps["netuid_to_min_stake"][netuid],
            "tempo": subnet_maps["netuid_to_tempo"][netuid],
            "trust_ratio": subnet_maps["netuid_to_trust_ratio"][netuid],
            "emission": subnet_maps["netuid_to_emission"][netuid],
            "max_weight_age": subnet_maps["netuid_to_max_weight_age"][netuid],
            "vote_mode": subnet_maps["netuid_to_vote_mode"][netuid]["vote_mode"],
            "adjustment_alpha": subnet_maps["netuid_to_adjustment_alpha"][netuid],
            "bonds_ma": subnet_maps["netuid_to_bonds_ma"].get(netuid, None),
            "maximum_set_weight_calls_per_epoch": subnet_maps["netuid_to_maximum_set_weight_calls_per_epoch"].get(netuid, 30),
            "target_registrations_per_interval": subnet_maps["netuid_to_target_registrations_per_interval"].get(netuid, default_target_registrations_per_interval),
            "target_registrations_interval": subnet_maps["netuid_to_target_registrations_interval"].get(netuid, default_target_registrations_interval),
            "max_registrations_per_interval": subnet_maps["netuid_to_max_registrations_per_interval"].get(netuid, default_max_registrations_per_interval),
        }

        result_subnets[netuid] = subnet

    return result_subnets


def get_global_params(c_client: CommuneClient) -> NetworkParams:
    """
    Returns global parameters of the whole commune ecosystem
    """

    query_all = c_client.query_batch(
        {
            "SubspaceModule": [
                ("MaxAllowedSubnets", []),
                ("MaxAllowedModules", []),
                ("MaxRegistrationsPerBlock", []),
                ("MaxNameLength", []),
                ("FloorDelegationFee", []),
                ("MaxAllowedWeightsGlobal", []),
                ("SubnetStakeThreshold", []),
                ("MinWeightStake", []),
                ("MinNameLength", []),
                ("BurnConfig", []),
                ("FloorFounderShare", []),
            ],
            "GovernanceModule": [
                ("GlobalGovernanceConfig", []),
                ("GeneralSubnetApplicationCost", []),
                ("Curator", []),
            ]
        }
    )
    governance_config: dict[str, int] = query_all["GlobalGovernanceConfig"] # type: ignore
    global_params: NetworkParams = {
        "max_allowed_subnets": int(query_all["MaxAllowedSubnets"]),
        "max_allowed_modules": int(query_all["MaxAllowedModules"]),
        "max_registrations_per_block": int(query_all["MaxRegistrationsPerBlock"]),
        "max_name_length": int(query_all["MaxNameLength"]),
        "min_burn": int(query_all["BurnConfig"]["min_burn"]), # type: ignore
        "max_burn": int(query_all["BurnConfig"]["max_burn"]), # type: ignore
        "min_weight_stake": int(query_all["MinWeightStake"]),
        "floor_delegation_fee": int(query_all["FloorDelegationFee"]),
        "max_allowed_weights": int(query_all["MaxAllowedWeightsGlobal"]),
        "curator": Ss58Address(query_all["Curator"]),
        "proposal_cost": int(governance_config["proposal_cost"]),
        "proposal_expiration": int(governance_config["proposal_expiration"]),
        "subnet_stake_threshold": int(query_all["SubnetStakeThreshold"]),
        "min_name_length": int(query_all["MinNameLength"]),
        "floor_founder_share": int(query_all["FloorFounderShare"]),
        "general_subnet_application_cost": int(query_all["GeneralSubnetApplicationCost"]),


    }
    return global_params


def concat_to_local_keys(
    balance: dict[str, int], local_key_info: dict[str, Ss58Address]
) -> dict[str, int]:
    key2: dict[str, int] = {
        key_name: balance.get(key_address, 0)
        for key_name, key_address in local_key_info.items()
    }

    return key2


def local_keys_to_freebalance(
    c_client: CommuneClient,
    local_keys: dict[str, Ss58Address],
) -> dict[str, int]:
    query_all = c_client.query_batch_map(
        {
            "System": [("Account", [])],
        }
    )
    balance_map = query_all["Account"]

    format_balances: dict[str, int] = {
        key: value["data"]["free"]
        for key, value in balance_map.items()
        if "data" in value and "free" in value["data"]
    }

    key2balance: dict[str, int] = concat_to_local_keys(format_balances, local_keys)

    return key2balance


def local_keys_to_stakedbalance(
    c_client: CommuneClient,
    local_keys: dict[str, Ss58Address],
    netuid: int = 0,
) -> dict[str, int]:
    query_all = c_client.query_batch_map(
        {
            "SubspaceModule": [("StakeTo", [netuid])],
        }
    )

    staketo_map = query_all["StakeTo"]

    format_stake: dict[str, int] = {
        key: sum(stake for _, stake in value) for key, value in staketo_map.items()
    }

    key2stake: dict[str, int] = concat_to_local_keys(format_stake, local_keys)

    return key2stake


def local_keys_allbalance(
    c_client: CommuneClient,
    local_keys: dict[str, Ss58Address],
    netuid: int | None = None,
) -> tuple[dict[str, int], dict[str, int]]:
    staketo_maps: list[Any] = []
    query_result = c_client.query_batch_map(
        {
            "SubspaceModule": [("SubnetNames", [])],
            "System": [("Account", [])],
        }
    )
    balance_map = query_result["Account"]
    all_netuids = list(query_result["SubnetNames"].keys())

    # update for all subnets
    netuids = all_netuids if netuid is None else [netuid]
    for uid in netuids:
        query_result = c_client.query_batch_map(
            {
                "SubspaceModule": [
                    ("StakeTo", [uid]),
                ],
            }
        )
        staketo_map = query_result.get("StakeTo", {})
        staketo_maps.append(staketo_map)

    format_balances: dict[str, int] = {
        key: value["data"]["free"]
        for key, value in balance_map.items()
        if "data" in value and "free" in value["data"]
    }

    key2balance: dict[str, int] = concat_to_local_keys(format_balances, local_keys)

    merged_staketo_map: dict[Any, Any] = {}

    # Iterate through each staketo_map in the staketo_maps list
    for staketo_map in staketo_maps:
        # Iterate through key-value pairs in the current staketo_map
        for key, value in staketo_map.items():
            # If the key is not present in the merged dictionary, add it
            if key not in merged_staketo_map:
                merged_staketo_map[key] = value
            else:
                # If the key exists, extend the existing list with the new values
                merged_staketo_map[key].extend(value)

    format_stake: dict[str, int] = {
        key: sum(stake for _, stake in value)
        for key, value in merged_staketo_map.items()
    }

    key2stake: dict[str, int] = concat_to_local_keys(format_stake, local_keys)

    key2balance = {
        k: v
        for k, v in sorted(key2balance.items(), key=lambda item: item[1], reverse=True)
    }

    key2stake = {
        k: v
        for k, v in sorted(key2stake.items(), key=lambda item: item[1], reverse=True)
    }

    return key2balance, key2stake



class ChainTransactionError(Exception):
    """Error for any chain transaction related errors."""


class NetworkError(Exception):
    """Base for any network related errors."""


class NetworkQueryError(NetworkError):
    """Network query related error."""


class NetworkTimeoutError(NetworkError):
    """Timeout error"""



"""
Data storage compatible with the *classic* `commune` library.
"""

# TODO: encryption

import base64
import hashlib
import json
import os.path
import time
from typing import Any

from nacl.secret import SecretBox
from nacl.utils import random

from communex.util import ensure_parent_dir_exists

# from cryptography.fernet import Fernet
# from cryptography.hazmat.primitives import hashes
# from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


COMMUNE_HOME = "~/.commune"
"""
COMMUNE_HOME

    classic commune data storage home directory.
"""


def _derive_key(password: str):
    # Derive a 256-bit key from the password using Blake2b
    key = hashlib.blake2b(password.encode(), digest_size=32).digest()
    return key


def _encrypt_data(password: str, data: Any) -> str:
    key = _derive_key(password)
    box = SecretBox(key)
    nonce = random(SecretBox.NONCE_SIZE)
    raw = json.dumps(data).encode()
    ciphertext = box.encrypt(raw, nonce).ciphertext
    encrypted = nonce + ciphertext
    decoded_data = base64.b64encode(encrypted).decode()
    return decoded_data


def _decrypt_data(password: str, data: str) -> Any:
    key = _derive_key(password)
    box = SecretBox(key)
    encrypted = base64.b64decode(data.encode())
    nonce = encrypted[:SecretBox.NONCE_SIZE]
    ciphertext = encrypted[SecretBox.NONCE_SIZE:]
    raw = box.decrypt(ciphertext, nonce)
    return json.loads(raw.decode())


def classic_load(path: str, mode: str = "json", password: str | None = None) -> Any:
    """
    Load data from commune data storage.

    Args:
        path: Data storage file path.
        mode: Data storage mode.

    Returns:
        Data loaded from the data storage.

    Todo:
        * Other serialization modes support. Only json mode is supported now.

    Raises:
        NotImplementedError: See Todo.
        AssertionError: Raised when the data is not in the classic format.
    """
    if mode != "json":
        raise NotImplementedError("Our commune data storage only supports json mode")

    full_path = os.path.expanduser(os.path.join(COMMUNE_HOME, path))
    with open(full_path, "r") as file:
        body = json.load(file)
        if body["encrypted"] and password is None:
            raise json.JSONDecodeError("Data is encrypted but no password provided", "", 0)
        if body["encrypted"] and password is not None:
            content = _decrypt_data(password, body["data"])
        else:
            content = body["data"]
    assert isinstance(body, dict)
    assert isinstance(body["timestamp"], int)
    assert isinstance(content, (dict, list, tuple, set, float, str, int))
    return content  # type: ignore


def classic_put(
    path: str, value: Any, mode: str = "json", password: str | None = None
):
    """
    Put data into commune data storage.

    Args:
        path: Data storage path.
        value: Data to store.
        mode: Data storage mode.
        encrypt: Whether to encrypt the data.

    Todo:
        * Encryption support.
        * Other serialization modes support. Only json mode is supported now.

    Raises:
        NotImplementedError: See Todo.
        TypeError: Raised when value is not a valid type.
        FileExistsError: Raised when the file already exists.
    """
    if mode != "json":
        raise NotImplementedError("Our commune data storage only supports json mode")
    if not isinstance(value, (dict, list, tuple, set, float, str, int)):
        raise TypeError(f"Invalid type for commune data storage value: {type(value)}")
    timestamp = int(time.time())

    full_path = os.path.expanduser(os.path.join(COMMUNE_HOME, path))

    if os.path.exists(full_path):
        raise FileExistsError(f"Commune data storage file already exists: {full_path}")

    ensure_parent_dir_exists(full_path)

    if password:
        value = _encrypt_data(password, value)
        encrypt = True
    else:
        encrypt = False

    with open(full_path, "w") as file:
        json.dump({'data': value, 'encrypted': encrypt, 'timestamp': timestamp},
                  file, indent=4)



from typing import Annotated, Optional

import typer

from communex import __version__

from ._common import ExtraCtxData
from .balance import balance_app
from .key import key_app
from .misc import misc_app
from .module import module_app
from .network import network_app
from .subnet import subnet_app

app = typer.Typer()

app.add_typer(key_app, name="key", help="Key operations")
app.add_typer(balance_app, name="balance", help="Balance operations")
app.add_typer(misc_app, name="misc", help="Other operations")
app.add_typer(module_app, name="module", help="Module operations")
app.add_typer(network_app, name="network", help="Network operations")
app.add_typer(subnet_app, name="subnet", help="Subnet operations")


def _version_callback(value: bool):
    if value:
        print(f"CommuneX {__version__}")
        raise typer.Exit()


def flag_option(
    flag: str,
    flag_envvar: str,
    flag_help: str,
    flag_short: str | None = None,
):
    flag_long = f"--{flag}"
    flag_short = f"-{flag[0]}" if flag_short is None else f"-{flag_short}"
    return typer.Option(
        flag_long,
        flag_short,
        is_flag=True,
        envvar=flag_envvar,
        help=flag_help
    )


@app.callback()
def main(
    ctx: typer.Context,
    json: Annotated[bool, flag_option(
        "json", "COMX_OUTPUT_JSON", "Output machine-readable JSON.")] = False,
    testnet: Annotated[bool, flag_option(
        "testnet", "COMX_USE_TESTNET", "Use testnet endpoints.")] = False,
    yes_to_all: Annotated[bool, flag_option(
        "yes", "COMX_YES_TO_ALL", "Say yes to all confirmation inputs.")] = False,
    version: Annotated[Optional[bool], typer.Option(
        callback=_version_callback)] = None,
):
    """
    CommuneX CLI {version}

    This command line interface is under development and subject to change.
    """
    # Pass the extra context data to the subcommands.
    ctx.obj = ExtraCtxData(
        output_json=json, use_testnet=testnet, yes_to_all=yes_to_all)


if main.__doc__ is not None:
    main.__doc__ = main.__doc__.format(version=__version__)



DECIMALS = 9


def from_nano(amount: int) -> float:
    """
    Converts from nano to j
    """

    return amount / (10**DECIMALS)


def to_nano(amount: float) -> int:
    """
    Converts from j to nano
    """

    return int(amount * (10**DECIMALS))


def from_horus(amount: int, subnet_tempo: int = 100) -> float:
    """
    Converts from horus to j
    """

    return amount / (10**DECIMALS * subnet_tempo)


def repr_j(amount: int):
    """
    Given an amount in nano, returns a representation of it in tokens/J.

    E.g. "103.2J".
    """

    return f"{from_nano(amount)}J"



import os.path
from typing import Any


def check_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def ensure_dir_exists(path: str) -> None:
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)


def ensure_parent_dir_exists(path: str) -> None:
    ensure_dir_exists(os.path.dirname(path))



from typing import Optional
import re
import typer
from typer import Context

from communex._common import BalanceUnit, format_balance, IPFS_REGEX
from communex.balance import to_nano
from communex.cli._common import (make_custom_context,
                                  print_table_from_plain_dict)
from communex.compat.key import (resolve_key_ss58_encrypted,
                                 try_classic_load_key)
from communex.errors import ChainTransactionError
from communex.faucet.powv2 import solve_for_difficulty_fast

balance_app = typer.Typer(no_args_is_help=True)


@balance_app.command()
def show(
    ctx: Context,
    key: str,
    unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None,
):
    """
    Gets the balances of a key.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)

    with context.progress_status(f"Getting balance of key {key_address}..."):
        subnets = client.query_map_subnet_names()
        netuids = list(subnets.keys())
        balance = client.get_balance(key_address)

        stakes: list[int] = []
        for uid in netuids:
            staketo = client.get_staketo(key_addr=key_address, netuid=uid)
            stakes.append(sum(staketo.values()))

        string_stakes = [format_balance(stake, unit) for stake in stakes]
        netuids = [str(uid) for uid in netuids]
        stake_dict = dict(zip(netuids, string_stakes))

        total = balance + sum(stakes)
        free, total = format_balance(balance, unit), format_balance(total, unit)

        print_table_from_plain_dict(stake_dict, ["Netuid", "Staked"], context.console)
        print_table_from_plain_dict(
            {"Free": free, "Total": total}, ["Result", "Amount"], context.console
        )


@balance_app.command()
def free_balance(
    ctx: Context,
    key: str,
    unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None,
):
    """
    Gets free balance of a key.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)

    with context.progress_status(f"Getting free balance of key {key_address}..."):
        balance = client.get_balance(key_address)

    context.output(format_balance(balance, unit))


@balance_app.command()
def staked_balance(
    ctx: Context,
    key: str,
    netuid: int = 0,
    unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None,
):
    """
    Gets the balance staked on the key itself.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)

    with context.progress_status(f"Getting staked balance of key {key_address}..."):
        staketo = client.get_staketo(key_addr=key_address, netuid=netuid)

    result = sum(staketo.values())

    context.output(format_balance(result, unit))


@balance_app.command()
def all_balance(
    ctx: Context,
    key: str,
    netuid: int = 0,
    unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None,
):
    """
    Gets entire balance of a key (free balance + staked balance).
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)

    with context.progress_status(f"Getting value of key {key_address}..."):
        staketo = client.get_staketo(key_address, netuid=netuid)
        stake_sum = sum(staketo.values())
        balance = client.get_balance(key_address)
        result = balance + stake_sum

    context.output(format_balance(result, unit))


@balance_app.command()
def get_staked(
    ctx: Context,
    key: str,
    netuid: int = 0,
    unit: BalanceUnit = BalanceUnit.joule,
    password: Optional[str] = None,
):
    """
    Gets total stake of a key it delegated across other keys.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    key_address = resolve_key_ss58_encrypted(key, context, password)

    with context.progress_status(f"Getting stake of {key_address}..."):
        result = client.get_stake(key=key_address, netuid=netuid)

    context.output(format_balance(result, unit))


@balance_app.command()
def transfer(ctx: Context, key: str, amount: float, dest: str):
    """
    Transfer amount to destination using key
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    nano_amount = to_nano(amount)
    resolved_key = try_classic_load_key(key, context)
    resolved_dest = resolve_key_ss58_encrypted(dest, context)

    if not context.confirm(
        f"Are you sure you want to transfer {amount} tokens to {dest}?"
    ):
        raise typer.Abort()

    with context.progress_status(f"Transferring {amount} tokens to {dest}..."):
        response = client.transfer(
            key=resolved_key, amount=nano_amount, dest=resolved_dest
        )

    if response.is_success:
        context.info(f"Transferred {amount} tokens to {dest}")
    else:
        raise ChainTransactionError(response.error_message)  # type: ignore


@balance_app.command()
def transfer_stake(
    ctx: Context, key: str, amount: float, from_key: str, dest: str, netuid: int = 0
):
    """
    Transfers stake of key from point A to point B
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    resolved_from = resolve_key_ss58_encrypted(from_key, context)
    resolved_dest = resolve_key_ss58_encrypted(dest, context)
    resolved_key = try_classic_load_key(key, context)
    nano_amount = to_nano(amount)

    with context.progress_status(
        f"Transferring {amount} tokens from {from_key} to {dest} on a subnet with netuid '{netuid}' ..."
    ):
        response = client.transfer_stake(
            key=resolved_key,
            amount=nano_amount,
            from_module_key=resolved_from,
            dest_module_address=resolved_dest,
            netuid=netuid,
        )

    if response.is_success:
        context.info(f"Transferred {amount} tokens from {from_key} to {dest}")
    else:
        raise ChainTransactionError(response.error_message)  # type: ignore


@balance_app.command()
def stake(
    ctx: Context,
    key: str,
    amount: float,
    dest: str,
    netuid: int = 0,
):
    """
    Stake amount to destination using key
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    nano_amount = to_nano(amount)
    resolved_key = try_classic_load_key(key, context)
    resolved_dest = resolve_key_ss58_encrypted(dest, context)
    delegating_message = (
        "By default you delegate DAO " 
        "voting power to the validator you stake to. "
        "In case you want to change this, call: "
        "`comx key power-delegation <key> --disable`."
    )
    context.info("INFO: ", style="bold green", end="") # type: ignore
    context.info(delegating_message) # type: ignore
    with context.progress_status(
        f"Staking {amount} tokens to {dest} on a subnet with netuid '{netuid}'..."
    ):
        response = client.stake(
            key=resolved_key, amount=nano_amount, dest=resolved_dest, netuid=netuid
        )

    if response.is_success:
        context.info(f"Staked {amount} tokens to {dest}")
    else:
        raise ChainTransactionError(response.error_message)  # type: ignore


@balance_app.command()
def unstake(ctx: Context, key: str, amount: float, dest: str, netuid: int = 0):
    """
    Unstake amount from destination using key
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    nano_amount = to_nano(amount)
    resolved_key = try_classic_load_key(key, context)
    resolved_dest = resolve_key_ss58_encrypted(dest, context)

    with context.progress_status(
        f"Unstaking {amount} tokens from {dest} on a subnet with netuid '{netuid}'..."
    ):
        response = client.unstake(
            key=resolved_key, amount=nano_amount, dest=resolved_dest, netuid=netuid
        )  # TODO: is it right?

    if response.is_success:
        context.info(f"Unstaked {amount} tokens from {dest}")
    else:
        raise ChainTransactionError(response.error_message)  # type: ignore


@balance_app.command()
def run_faucet(
    ctx: Context,
    key: str,
    num_processes: Optional[int] = None,
    num_executions: int = 1,
):
    context = make_custom_context(ctx)
    use_testnet = ctx.obj.use_testnet
    if not use_testnet:
        context.error("Faucet only enabled on testnet")
        return
    resolved_key = try_classic_load_key(key, context)
    client = context.com_client()
    for _ in range(num_executions):
        with context.progress_status("Solving PoW..."):
            solution = solve_for_difficulty_fast(
                client,
                resolved_key,
                client.url,
                num_processes=num_processes,
            )
        with context.progress_status("Sending solution to blockchain"):
            params = {
                "block_number": solution.block_number,
                "nonce": solution.nonce,
                "work": solution.seal,
                "key": resolved_key.ss58_address,
            }
            client.compose_call("faucet", params=params, unsigned=True, module="FaucetModule", key=resolved_key.ss58_address)  # type: ignore


@balance_app.command()
def transfer_dao_funds(
    ctx: Context,
    signer_key: str,
    amount: float,
    cid_hash: str,
    dest: str,
):
    context = make_custom_context(ctx)
    if not re.match(IPFS_REGEX, cid_hash):
        context.error(f"CID provided is invalid: {cid_hash}")
        exit(1)
    ipfs_prefix = "ipfs://"
    cid = ipfs_prefix + cid_hash

    client = context.com_client()

    nano_amount = to_nano(amount)
    dest = resolve_key_ss58_encrypted(dest, context)
    signer_keypair = try_classic_load_key(signer_key, context)
    client.add_transfer_dao_treasury_proposal(
        signer_keypair, cid, nano_amount, dest
    )


"""
The CommuneX library package.

Submodules:
    * `communex.client`: A lightweigh yet faster client for the Commune network.
    * `.compat`: Compatibility layer for the *classic* `commune` library.
    * `.types`: CommuneX common types.
    * `.key`: Key related functions.

.. include:: ../../README.md
"""

import importlib.metadata

if not __package__:
    __version__ = "0.0.0"
else:
    __version__ = importlib.metadata.version(__package__)



import ipaddress
import json
import os.path
import re
from typing import Any, Callable, Generic, Optional, Protocol, TypeVar

import requests


def check_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def ensure_dir_exists(path: str) -> None:
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)


def ensure_parent_dir_exists(path: str) -> None:
    ensure_dir_exists(os.path.dirname(path))


def bytes_to_hex(value: str | bytes) -> str:
    """
    Converts a string or bytes object to its hexadecimal representation.

    If the input is already a string, it assumes that the string is already in
    hexadecimal format and returns it as is. If the input is bytes, it converts
    the bytes to their hexadecimal string representation.

    Args:
        x: The input string or bytes object to be converted to hexadecimal.

    Returns:
        The hexadecimal representation of the input.

    Examples:
        _to_hex(b'hello') returns '68656c6c6f'
        _to_hex('68656c6c6f') returns '68656c6c6f'
    """
    if isinstance(value, bytes):
        return value.hex()
    assert isinstance(value, str)
    # TODO: Check that `value` is a valid hexadecimal string
    return value


def is_ip_valid(ip: str) -> bool:
    """
    Checks if an ip address is valid
    """
    try:
        ipaddress.ip_address(ip)
        return True
    except ValueError:
        return False


T = TypeVar("T")


class SetterGetterFn(Generic[T], Protocol):
    def __call__(self, x: T = ..., /) -> T:
        ...


def create_state_fn(default: Callable[..., T]) -> SetterGetterFn[T]:
    """
    Creates a state function that can be used to get or set a value.
    """
    value = default()

    def state_function(input: Optional[T] = None):
        nonlocal value
        if input is not None:
            value = input
        return value

    return state_function


def get_json_from_cid(cid: str) -> dict[Any, Any] | None:
    gateway = "https://ipfs.io/ipfs/"
    try:
        result = requests.get(gateway + cid)
        if result.ok:
            return result.json()
        return None
    except Exception:
        return None


def convert_cid_on_proposal(proposals: dict[int, dict[str, Any]]):
    unwrapped: dict[int, dict[str, Any]] = {}
    for prop_id, proposal in proposals.items():
        data = proposal.get("data")
        if data and "Custom" in data:
            metadata = proposal["metadata"]
            cid = metadata.split("ipfs://")[-1]
            queried_cid = get_json_from_cid(cid)
            if queried_cid:
                body = queried_cid.get("body")
                if body:
                    try:
                        queried_cid["body"] = json.loads(body)
                    except Exception:
                        pass
            proposal["Custom"] = queried_cid
        unwrapped[prop_id] = proposal
    return unwrapped


HEX_PATTERN = re.compile(r"^[0-9a-fA-F]+$")


# TODO: merge `is_hex_string` into `parse_hex`
def is_hex_string(string: str):
    return bool(HEX_PATTERN.match(string))


def parse_hex(hex_str: str) -> bytes:
    if hex_str[0:2] == "0x":
        return bytes.fromhex(hex_str[2:])
    else:
        return bytes.fromhex(hex_str)



import importlib.util
from typing import Any, Optional, cast

import typer
import uvicorn
from typer import Context

import communex.balance as c_balance
from communex._common import intersection_update
from communex.cli._common import (make_custom_context, print_module_info,
                                  print_table_from_plain_dict)
from communex.compat.key import try_classic_load_key
from communex.errors import ChainTransactionError
from communex.misc import get_map_modules
from communex.module._rate_limiters.limiters import (IpLimiterParams,
                                                     StakeLimiterParams)
from communex.module.server import ModuleServer
from communex.util import is_ip_valid

from communex.key import check_ss58_address
from communex.types import Ss58Address


module_app = typer.Typer(no_args_is_help=True)


def list_to_ss58(str_list: list[str] | None) -> list[Ss58Address] | None:
    """Raises AssertionError if some input is not a valid Ss58Address."""
    
    if str_list is None:
        return None
    new_list: list[Ss58Address] = []
    for item in str_list:
        new_item = check_ss58_address(item)
        new_list.append(new_item)
    return new_list


# TODO: refactor module register CLI
# - module address should be a single (arbitrary) parameter
# - key can be infered from name or vice-versa?
@module_app.command()
def register(
    ctx: Context,
    name: str,
    key: str,
    ip: Optional[str] = None,
    port: Optional[int] = None,
    netuid: Optional[int] = None,
    stake: Optional[float] = None,
    metadata: Optional[str] = None,
    new_subnet_name: Optional[str] = None,
):
    """
    Registers a module on the Commune network.
    """
    context = make_custom_context(ctx)
    client = context.com_client()
    if metadata and len(metadata) > 59:
        raise ValueError("Metadata must be less than 60 characters")

    match (netuid, new_subnet_name):
        case (None, None):
            raise ValueError(
                "`--netuid` or `--new_subnet_name` argument must be passed into the CLI command \n"
                "f.e comx module register <name> <ip> <port> <key> --netuid <netuid>"
            )
        case (netuid, None):
            assert netuid is not None
            subnet_name = client.get_subnet_name(netuid)
            burn = client.get_burn(netuid)
        case (None, new_subnet_name):
            subnet_name = new_subnet_name
            burn = client.get_min_burn()
        case (_, _):
            raise ValueError(
                "`--netuid` and `--new_subnet_name` cannot be provided at the same time"
            )

    do_burn = context.confirm(
        f"{c_balance.from_nano(burn)} $COMAI will be permanently burned. Do you want to continue?"
    )

    if not do_burn:
        print("Not registering")
        raise typer.Abort()

    with context.progress_status(f"Registering Module {name}..."):
        if stake is not None:
            stake_nano = c_balance.to_nano(stake)
        else:
            min_stake = client.get_min_stake(netuid) if netuid is not None else 0
            stake_nano = (
                min_stake + burn + c_balance.to_nano(0.1)
            )  # 0.1 extra needed for the call to succeed

        resolved_key = try_classic_load_key(key, context)

        address = f"{ip}:{port}"

        response = client.register_module(
            resolved_key,
            name=name,
            address=address,
            subnet=subnet_name,
            min_stake=stake_nano,
            metadata=metadata,
        )

        if response.is_success:
            context.info(f"Module {name} registered")
        else:
            raise ChainTransactionError(response.error_message)  # type: ignore


@module_app.command()
def update(
    ctx: Context,
    key: str,
    name: Optional[str] = None,
    ip: Optional[str] = None,
    port: Optional[int] = None,
    delegation_fee: int = 20,
    netuid: int = 0,
    metadata: Optional[str] = None,
):
    """
    Update module with custom parameters.
    """

    context = make_custom_context(ctx)
    client = context.com_client()
    if metadata and len(metadata) > 59:
        raise ValueError("Metadata must be less than 60 characters")
    resolved_key = try_classic_load_key(key)

    if ip and not is_ip_valid(ip):
        raise ValueError("Invalid ip address")
    modules = get_map_modules(client, netuid=netuid, include_balances=False)
    modules_to_list = [value for _, value in modules.items()]

    module = next(
        (
            item for item in modules_to_list if item["key"] == resolved_key.ss58_address
        ),
        None
    )

    if module is None:
        raise ValueError(f"Module {name} not found")
    module_params = {
        "name": name,
        "ip": ip,
        "port": port,
        "delegation_fee": delegation_fee,
        "metadata": metadata,
    }
    to_update = {
        key: value for key, value in module_params.items() if value is not None
    }
    current_ip, current_port = module["address"].split(":")
    new_ip = to_update.get("ip", current_ip)
    new_port = to_update.get("port", current_port)
    address = f"{new_ip}:{new_port}"
    to_update["address"] = address
    updated_module = intersection_update(dict(module), to_update)
    module.update(updated_module)  # type: ignore
    with context.progress_status(
        f"Updating Module on a subnet with netuid '{netuid}' ..."
    ):
        response = client.update_module(
            key=resolved_key,
            name=module["name"],
            address=module["address"],
            delegation_fee=module["delegation_fee"],
            netuid=netuid,
            metadata=module["metadata"],
        )

    if response.is_success:
        context.info(f"Module {key} updated")
    else:
        raise ChainTransactionError(response.error_message)  # type: ignore


@module_app.command()
def serve(
    ctx: typer.Context,
    class_path: str,
    key: str,
    port: int = 8000,
    ip: Optional[str] = None,
    subnets_whitelist: Optional[list[int]] = [0],
    whitelist: Optional[list[str]] = None,
    blacklist: Optional[list[str]] = None,
    ip_blacklist: Optional[list[str]] = None,
    test_mode: Optional[bool] = False,
    request_staleness: int = typer.Option(120),
    use_ip_limiter: Optional[bool] = typer.Option(
        False, help=("If this value is passed, the ip limiter will be used")
    ),
    token_refill_rate_base_multiplier: Optional[int] = typer.Option(
        None, help=(
            "Multiply the base limit per stake. Only used in stake limiter mode."
        )
    )
):
    """
    Serves a module on `127.0.0.1` on port `port`. `class_path` should specify
    the dotted path to the module class e.g. `module.submodule.ClassName`.
    """
    context = make_custom_context(ctx)
    use_testnet = context.get_use_testnet()
    path_parts = class_path.split(".")
    match path_parts:
        case [*module_parts, class_name]:
            module_path = ".".join(module_parts)
            if not module_path:
                # This could do some kind of relative import somehow?
                raise ValueError(
                    f"Invalid class path: `{class_path}`, module name is missing"
                )
            if not class_name:
                raise ValueError(
                    f"Invalid class path: `{class_path}`, class name is missing"
                )
        case _:
            # This is impossible
            raise Exception(f"Invalid class path: `{class_path}`")

    try:
        module = importlib.import_module(module_path)
    except ModuleNotFoundError:
        context.error(f"Module `{module_path}` not found")
        raise typer.Exit(code=1)

    try:
        class_obj = getattr(module, class_name)
    except AttributeError:
        context.error(f"Class `{class_name}` not found in module `{module}`")
        raise typer.Exit(code=1)

    keypair = try_classic_load_key(key, context)
    if test_mode:
        subnets_whitelist = None
    token_refill_rate = token_refill_rate_base_multiplier or 1
    limiter_params = IpLimiterParams() if use_ip_limiter else StakeLimiterParams(token_ratio=token_refill_rate)

    if whitelist is None:
        context.info(
            "WARNING: No whitelist provided, will accept calls from any key"
        )



    try:
        whitelist_ss58 = list_to_ss58(whitelist)
    except AssertionError:
        context.error(
            "Invalid SS58 address passed to whitelist"
        )
        exit(1)
    try:
        blacklist_ss58 = list_to_ss58(blacklist)
    except AssertionError:
        context.error(
            "Invalid SS58 address passed on blacklist"
        )
        exit(1)
    cast(list[Ss58Address] | None, whitelist)
    
    server = ModuleServer(
        class_obj(), keypair, 
        whitelist=whitelist_ss58, blacklist=blacklist_ss58, 
        subnets_whitelist=subnets_whitelist, 
        max_request_staleness=request_staleness,
        limiter=limiter_params,
        ip_blacklist=ip_blacklist,
        use_testnet=use_testnet
    )
    app = server.get_fastapi_app()
    host = ip or "127.0.0.1"
    uvicorn.run(app, host=host, port=port)  # type: ignore


@module_app.command()
def info(ctx: Context, name: str, balance: bool = False, netuid: int = 0):
    """
    Gets module info
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    with context.progress_status(
        f"Getting Module {name} on a subnet with netuid {netuid}…"
    ):
        modules = get_map_modules(client, netuid=netuid, include_balances=balance)
        modules_to_list = [value for _, value in modules.items()]

        module = next((item for item in modules_to_list if item["name"] == name), None)

    if module is None:
        raise ValueError("Module not found")

    general_module = cast(dict[str, Any], module)
    print_table_from_plain_dict(general_module, ["Params", "Values"], context.console)


@module_app.command(name="list")
def inventory(ctx: Context, balances: bool = False, netuid: int = 0):
    """
    Modules stats on the network.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    # with context.progress_status(
    #     f"Getting Modules on a subnet with netuid {netuid}..."
    # ):
    modules = cast(dict[str, Any], get_map_modules(
        client, netuid=netuid, include_balances=balances))

    # Convert the values to a human readable format
    modules_to_list = [value for _, value in modules.items()]

    miners: list[Any] = []
    validators: list[Any] = []
    inactive: list[Any] = []

    for module in modules_to_list:
        if module["incentive"] == module["dividends"] == 0:
            inactive.append(module)
        elif module["incentive"] > module["dividends"]:
            miners.append(module)
        else:
            validators.append(module)

    print_module_info(client, miners, context.console, netuid, "miners")
    print_module_info(client, validators, context.console, netuid, "validators")
    print_module_info(client, inactive, context.console, netuid, "inactive")



import re
from typing import Any, Optional, cast

import typer
from rich.progress import track
from typer import Context

import communex.balance as c_balance
from communex.cli._common import (CustomCtx, make_custom_context,
                                  print_table_from_plain_dict)
from communex.client import CommuneClient
from communex.compat.key import (
    local_key_addresses, 
    try_classic_load_key
    )
from communex.misc import (IPFS_REGEX, get_global_params,
                           local_keys_to_stakedbalance)
from communex.types import NetworkParams
from communex.util import convert_cid_on_proposal

network_app = typer.Typer(no_args_is_help=True)


@network_app.command()
def last_block(ctx: Context, hash: bool = False):
    """
    Gets the last block
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    info = "number" if not hash else "hash"

    block = client.get_block()
    block_info = None
    if block:
        block_info = block["header"][info]

    context.output(str(block_info))


@network_app.command()
def params(ctx: Context):
    """
    Gets global params
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    with context.progress_status("Getting global network params ..."):
        global_params = get_global_params(client)

    general_params: dict[str, Any] = cast(dict[str, Any], global_params)
    print_table_from_plain_dict(
        general_params, ["Global params", "Value"], context.console
    )


@network_app.command()
def list_proposals(ctx: Context, query_cid: bool = typer.Option(True)):
    """
    Gets proposals
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    with context.progress_status("Getting proposals..."):
        try:
            proposals = client.query_map_proposals()
            if query_cid:
                proposals = convert_cid_on_proposal(proposals)
        except IndexError:
            context.info("No proposals found.")
            return

    for proposal_id, batch_proposal in proposals.items():
        status = batch_proposal["status"]
        if isinstance(status, dict):
            batch_proposal["status"] = [*status.keys()][0]
        print_table_from_plain_dict(    
            batch_proposal, [
                f"Proposal id: {proposal_id}", "Params"], context.console
        )


@network_app.command()
def propose_globally(
    ctx: Context,
    key: str,
    cid: str,
    max_allowed_modules: int = typer.Option(None),
    max_registrations_per_block: int = typer.Option(None),
    max_name_length: int = typer.Option(None),
    min_name_length: int = typer.Option(None),
    min_burn: int = typer.Option(None),
    max_burn: int = typer.Option(None),
    min_weight_stake: int = typer.Option(None),
    max_allowed_subnets: int = typer.Option(None),
    curator: str = typer.Option(None),
    proposal_cost: int = typer.Option(None),
    proposal_expiration: int = typer.Option(None),
    subnet_stake_threshold: int = typer.Option(None),
    general_subnet_application_cost: int = typer.Option(None),
    floor_founder_share: int = typer.Option(None),
    floor_delegation_fee: int = typer.Option(None),
    max_allowed_weights: int = typer.Option(None),
):
    provided_params = locals().copy()
    provided_params.pop("ctx")
    provided_params.pop("key")
    provided_params = {
        key: value for key, value in provided_params.items() if value is not None
    }
    """
    Adds a global proposal to the network.
    """
    context = make_custom_context(ctx)
    resolved_key = try_classic_load_key(key, context)
    client = context.com_client()

    provided_params = cast(NetworkParams, provided_params)
    global_params = get_global_params(client)
    global_params.update(provided_params)
    
    if not re.match(IPFS_REGEX, cid):
        context.error(f"CID provided is invalid: {cid}")
        exit(1)
    with context.progress_status("Adding a proposal..."):
        client.add_global_proposal(resolved_key, global_params, cid)


def get_valid_voting_keys(
    ctx: CustomCtx,
    client: CommuneClient,
    proposal: dict[str, Any],
) -> dict[str, int]:
    local_keys = local_key_addresses(ctx=ctx, universal_password=None)

    if proposal.get("SubnetParams"):
        proposal_netuid = proposal["SubnetParams"]["netuid"]
        assert (isinstance(proposal_netuid, int))
        keys_stake = local_keys_to_stakedbalance(client, local_keys, netuid=proposal_netuid)
    else:
        keys_stake: dict[str, int] = {}
        subnets = client.query_map_subnet_names()
        for netuid in track(subnets.keys(), description="Checking valid keys..."):
            subnet_stake = local_keys_to_stakedbalance(client, local_keys, netuid=netuid)
            keys_stake = {
                key: keys_stake.get(key, 0) + subnet_stake.get(key, 0)
                for key in set(keys_stake) | set(subnet_stake)
            }
    keys_stake = {key: stake for key,
                  stake in keys_stake.items() if stake >= 5}
    return keys_stake


@network_app.command()
def vote_proposal(
    ctx: Context,
    proposal_id: int,
    key: Optional[str] = None,
    agree: bool = typer.Option(True, "--disagree"),
):
    """
    Casts a vote on a specified proposal. Without specifying a key, all keys on disk will be used.
    """
    context = make_custom_context(ctx)
    client = context.com_client()
    proposals = client.query_map_proposals()
    proposal = proposals[proposal_id]

    if key is None:
        context.info("Voting with all keys on disk...")
        delegators = client.get_voting_power_delegators()
        keys_stake = get_valid_voting_keys(context, client, proposal)
        keys_stake = {
            key: stake for key,
            stake in keys_stake.items() if key not in delegators
            }
    else:
        keys_stake = {key: None}

    for voting_key in track(keys_stake.keys(), description="Voting..."):
        
        resolved_key = try_classic_load_key(voting_key, context)
        try:
            client.vote_on_proposal(resolved_key, proposal_id, agree)
        except Exception as e:
            print(f"Error while voting with key {key}: ", e)
            print("Skipping...")
            continue


@network_app.command()
def unvote_proposal(ctx: Context, key: str, proposal_id: int):
    """
    Retracts a previously cast vote on a specified proposal.
    """
    context = make_custom_context(ctx)
    client = context.com_client()

    resolved_key = try_classic_load_key(key, context)
    with context.progress_status(f"Unvoting on a proposal {proposal_id}..."):
        client.unvote_on_proposal(resolved_key, proposal_id)


@network_app.command()
def add_custom_proposal(ctx: Context, key: str, cid: str):
    """
    Adds a custom proposal.
    """
    context = make_custom_context(ctx)
    if not re.match(IPFS_REGEX, cid):
        context.error(f"CID provided is invalid: {cid}")
        exit(1)
    else:
        ipfs_prefix = "ipfs://"
        cid = ipfs_prefix + cid
    client = context.com_client()
    # append the ipfs hash
    ipfs_prefix = "ipfs://"
    cid = ipfs_prefix + cid

    resolved_key = try_classic_load_key(key, context)

    with context.progress_status("Adding a proposal..."):
        client.add_custom_proposal(resolved_key, cid)


@network_app.command()
def registration_burn(
    ctx: Context,
    netuid: int,
):
    """
    Appraises the cost of registering a module on the Commune network.
    """

    context = make_custom_context(ctx)
    client = context.com_client()

    burn = client.get_burn(netuid)
    registration_cost = c_balance.from_nano(burn)
    context.info(
        f"The cost to register on a netuid: {netuid} is: {registration_cost} $COMAI"
    )

